from datetime import datetime
from sys import stderr


class Rectangle:
    """
    Rectangle class represents a rectangle that is able to compute its area.
    """

    def __init__(self, width, height):
        self.width = width
        self.height = height

    def get_area(self):
        return self.height * self.width

    def set_size(self, new_width, new_height):
        self.width = new_width
        self.height = new_height


class Matrix:
    """
    Matrix implementation using the Compressed sparse row (CSR, CRS, or Yale format) method.
    It represents a matrix M by three (one-dimensional) arrays, that respectively contain
    nonzero values, the extents of rows, and column indices. This format allows fast row
    access and matrix-vector multiplications (Mx). It is not intended for modifications,
    since it may imply a lot of operations.

    For more information see: https://en.wikipedia.org/wiki/Sparse_matrix

    Doctests:
    >>> m = Matrix([[0, 0, 0, 0], [5, 8, 0, 0], [0, 0, 3, 0], [0, 6, 0, 0]])
    >>> m.print_repr()
    [5, 8, 3, 6]
    [0, 1, 2, 1]
    [0, 0, 2, 3, 4]

    >>> m.get_data(3, 1)
    6
    """

    def __init__(self, matrix):
        """
        Creates the matrix representation
        :param matrix: list of lists representing list of rows of ints of the same length
        """
        self.height = len(matrix)
        self.width = len(matrix[0])
        self.values = list()
        self.col_index = list()
        self.row_index = list()

        self.row_index.append(0)

        last_row_idx = 0

        for row in matrix:
            non_zeroes = 0
            for idx in range(len(row)):
                if row[idx] != 0:
                    non_zeroes += 1
                    self.values.append(row[idx])
                    self.col_index.append(idx)

            self.row_index.append(non_zeroes + last_row_idx)
            last_row_idx += non_zeroes

    def get_data(self, row, column):
        """
        Gets the number at the specified position.
        If the position is outside the matrix, zero is returned.

        :param row: The row of the requested number
        :param column: The column of the requested number
        :return: The requested number in the matrix
        """
        if len(self.row_index) <= row or (
                self.row_index[row + 1] - self.row_index[row]) == 0:  # no data in the representation -> zero is present
            return 0
        else:
            for i in range(self.row_index[row], self.row_index[row + 1]):  # non-zero data present
                if self.col_index[i] == column:  # the requested column is non-zero, return the value
                    return self.values[i]

            return 0

    def sum_with_another(self, another):
        """
        Sums two matrices together. They have to have the same sizes.
        :param another: The other matrix to be "added" to this one (self)
        :return: New matrix instance as the sum of the two matrices
        """
        raw_rows = list()

        if self.width != another.width or self.height != another.height:
            raise ValueError("Matrices have to have the same size")

        for i in range(self.height):
            row = list()
            for j in range(self.width):
                row.append(self.get_data(i, j) + another.get_data(i, j))
            raw_rows.append(row)

        return Matrix(raw_rows)

    def product_with_another(self, another):
        """
        Computes produc of two matrices together. They have to have the correct sizes, i.e.,
        the width of the first one has to be the same as the height of the second one and vice versa.
        :param another: The other matrix
        :return: New matrix instance as the prodict of the two matrices
        """

        def compute_item(x, y):
            """
            Helper method copmuting a single cell of the product
            :param x: the x position -- the line index
            :param y: the y position -- the column index
            :return:
            """
            nonlocal self
            nonlocal another

            result = 0
            for k in range(self.width):
                result += self.get_data(x, k) * another.get_data(k, y)

            return result

        raw_rows = list()

        if self.width != another.height or self.height != another.width:
            raise ValueError("Matrices have to have the correct size")

        for i in range(self.height):
            row = list()
            for j in range(self.width):
                row.append(compute_item(i, j))

            raw_rows.append(row)

        return Matrix(raw_rows)

    def print_repr(self):
        print(self.values)
        print(self.col_index)
        print(self.row_index)


class Logger:
    """
    Logger class provides the logging funcionality. It maintains a level, and prints out a message if and
    only if the logger level is less than or equal to the level provided as an argument to the print method.
    It also maintains a list of printers and prints each message by all the printers.
    """

    def __init__(self, name):
        self.printers = list()
        self.name = name
        self.level = 0
        self.formatter = None

    def set_level(self, level):
        self.level = level

    def log(self, level, message):
        if self.level <= level:
            if self.formatter is not None:
                message = self.formatter.format(self, message)

            for printer in self.printers:
                printer.print(message)

    def add_printer(self, printer):
        self.printers.append(printer)

    def set_formatter(self, formatter):
        self.formatter = formatter


class BasicPrinter:
    """
    The basic printer, just prints messages to standard output
    """

    @staticmethod
    def print(message):
        print(message)


class FilePrinter:
    """
    Prints a timestamp in front of each message, prints to stdout
    """

    def __init__(self, filename):
        self.filename = filename

    def print(self, message):
        with open(self.filename, mode='at') as f:
            print(message, file=f)


class ErrorPrinter:
    """
    Prints messages to stderr
    """

    @staticmethod
    def print(message):
        print(message, file=stderr)


class NameFormatter:
    """
    Extends the message by the name of the logger
    """

    @staticmethod
    def format(logger, message):
        return logger.name + ': ' + message


class TimestampFormatter:
    """
    Extends the message by a timestamp
    """

    @staticmethod
    def format(logger, message):
        return str(datetime.now()) + ': ' + message

### rectangle
r = Rectangle(4, 5)
print(r.get_area())
r.set_size(2, 3)
print(r.get_area())

m = Matrix([[0, 0, 0, 0], [5, 8, 0, 0], [0, 0, 3, 0], [0, 6, 0, 0]])
m.print_repr()
print(m.get_data(3, 1))

result = m.sum_with_another(m)
print(result.get_data(3, 1))

m2 = Matrix([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]])
result = m2.product_with_another(m)
result.print_repr()


logger = Logger("The Logger")
logger.add_printer(BasicPrinter())
logger.add_printer(FilePrinter('output.log'))
logger.set_level(2)
logger.set_formatter(TimestampFormatter())
logger.log(2, "Logged message")
logger.log(0, "Message ignored due to loglevel")
